// Name	::	Load Add-On
// ================
// Made by	::	McTwist
// Date	::	10-01-13
// Info	::	Manually loads an add-on without restart server
// License	::	Free to use
// ================

$AddOns_Amount = 0;
$AddOns_Path = "Add-Ons/*_*/server.cs"; // Only for add-ons that can be executed on a server

// Loading an add-on
function ServerCmdLoadAddOn(%client, %add, %next)
{
	// Host only
	if (!%client.isHost())
		return;
	
	// Does not want to exceed this
	if ($datablockExceededCount > 0)
	{
		messageClient(%client, 'MessageBoxOk', "Exceed Datablocks", "You are unable to load more add-ons.");
		return;
	}
	
	messageAll('', "Trying load add-on \c6" @ %add);
	
	// Load
	%load = LoadAddOn(%add, 0);
	if (%load == 0 && !%next)
		UpdateClientDatablocks();
	
	if (!%next)
	{
		if (%load == 2)
			messageAll('', "\c6" @ %add @ "\c0 does not exist");
		else if (%load == 4)
			messageAll('', "\c6" @ %add @ "\c0 is not a valid add-on");
		else if (%load == 16)
			messageAll('', "\c6" @ %add @ "\c0 already loaded");
		else if (%load == 32)
			messageAll('', "Failed reloading \c6" @ %add @ "\c0");
		else if (%load == 64)
			messageAll('', "\c6" @ %add @ "\c0 is on the crap-on list");
		else
		{
			messageAll('', "Add-On loaded.");
			// Repopulate list
			schedule(0, 0, PopulateAddOnList);
		}
	}
}

// Reloading an add-on
function ServerCmdReLoadAddOn(%client, %add, %next)
{
	// Host only
	if (!%client.isHost())
		return;
	
	// Does not want to exceed this
	if ($datablockExceededCount > 0)
	{
		messageClient(%client, 'MessageBoxOk', "Exceed Datablocks", "You are unable to load more add-ons.");
		return;
	}
	
	messageAll('', "Trying reload add-on \c6" @ %add);
	
	%load = LoadAddOn(%add, 1);
	if (%load == 0 && !%next)
		UpdateClientDatablocks();
	
	if (!%next)
	{
		if (%load == 2)
			messageAll('', "\c6" @ %add @ "\c0 does not exist");
		else if (%load == 4)
			messageAll('', "\c6" @ %add @ "\c0 is not a valid add-on");
		else if (%load == 32)
			messageAll('', "Failed reloading \c6" @ %add @ "\c0");
		else if (%load == 64)
			messageAll('', "\c6" @ %add @ "\c0 is on the crap-on list");
		else
			messageAll('', "Add-On loaded.");
	}
}

// Get add-ons from server
function ServerCmdGetAddOns(%client)
{
	// Host only
	if (!%client.isHost())
	{
		messageClient(%client, '', "Host only!");
		return;
	}
	commandToClient(%client, 'ClearAddOnList');
	
	// Send list
	for (%i = 1; %i <= $AddOns_Amount; %i += 17)
		commandToClient(%client, 'ReceiveAddOns', $AddOns_Num_[%i], $AddOns_Num_[%i+1], $AddOns_Num_[%i+2], $AddOns_Num_[%i+3], $AddOns_Num_[%i+4], $AddOns_Num_[%i+5], $AddOns_Num_[%i+6], $AddOns_Num_[%i+7], $AddOns_Num_[%i+8], $AddOns_Num_[%i+9], $AddOns_Num_[%i+10], $AddOns_Num_[%i+11], $AddOns_Num_[%i+12], $AddOns_Num_[%i+13], $AddOns_Num_[%i+14], $AddOns_Num_[%i+15], $AddOns_Num_[%i+16]);
}

// Load an add-on
function LoadAddOn(%add, %force)
{
	// Allways discover file
	if (!discoverfile("Add-Ons/" @ %add @ ".zip"))
	{
		echo("Failed loading \"" @ %add @ "\".");
		return 2;
	}
	$AddOns_CRC_[%add] = getFileCRC("Add-Ons/" @ %add @ ".zip");
	
	// Verify it
	if ((%code = VerifyAddOn(%add)) > 0)
	{
		// Do something about it, or just ignore
		switch (%code)
		{
			case 1: warn("Have not loaded server."); return 1;
			case 2: warn("\"" @ %add @ "\" does not exist."); return 2;
			case 4: warn("\"" @ %add @ "\" is not a valid add-on."); return 4;
			case 8: return 8;
			case 16:
				if (!%force)
				{
					echo("\"" @ %add @ "\" already loaded.");
					return 16;
				}
			case 32:
				echo("\"" @ %add @ "\" has been changed. Reloads...");
				// Find the file again
				if (!discoverfile("Add-Ons/" @ %add @ ".zip"))
				{
					echo("Failed reloading \"" @ %add @ "\".");
					return 32;
				}
			case 64:
				echo("\"" @ %add @ "\" is on the crap-on list.");
				return 64;
		}
		echo("Reloading \"" @ %add @ "\".");
	}
	
	// Mark as loaded
	$AddOnLoaded__[%add] = 1;
	// Remove from list
	if ($AddOns_Name_[%add] >= 1)
	{
		$AddOns_Num_[$AddOns_Name_[%add]] = "";
		$AddOns_Name_[%add] = "";
	}
	// Load the add-on
	exec("Add-Ons/" @ %add @ "/server.cs");
	return 0;
}

// Check if add-on can load
// Returns 0 if correct verified, else returning a positive number code
function VerifyAddOn(%add)
{
	%file = "Add-Ons/" @ %add;
	
	// No need to continue
	if (!isFunction("isValidAddOn"))
		return 1;
	// Does it even exist
	if (!isFile(%file @ ".zip"))
		return 2;
	// Validate
	if (!(isFile(%file @ "/server.cs") && isValidAddOn(%add)))
		return 4;
	// Verify the scripts
	if (!VerifyAddOnScripts(%add))
		return 8;
	// Already loaded
	if ($AddOnLoaded__[%add] == 1)
		return 16;
	// It hasn't been updated correctly
	%crc = getFileCRC(%file @ ".zip");
	if ($AddOns_CRC_[%add] !$= %crc)
		return 32;
	// In the crap on list
	if ($CrapOnCRC_[%crc] || $CrapOnDedicatedName_[%add] || $CrapOnName_[%add])
		return 64;
	return 0;
}

// Update datablocks for each client
function UpdateClientDatablocks()
{
	// Magically make the datablocks visible
	for (%i = 0; %i < ClientGroup.getCount(); %i++)
	{
		%client = ClientGroup.getObject(%i);
		// Cheating
		%client.reloadDatablocks = true;
		// Transmit datablocks
		%client.transmitDataBlocks($missionSequence);
	}
}

// First instance when searching for add-ons
function PopulateAddOnList()
{
	// Clear the list
	deleteVariables("$AddOns_Num_*");
	deleteVariables("$AddOns_Name_*");
	$AddOns_Amount = 0;
	// Find first file
	if (strlen(%filepath = findFirstFile($AddOns_Path)) == 0)
	{
		echo("None add-on found.");
		return;
	}
	AddAddOnToList(%filepath);
	schedule(0, 0, "PopulateAddOnList_Schedule");
}

// The rest when searching for add-ons
function PopulateAddOnList_Schedule()
{
	for (%i = 0; %i < 10 && strlen(%filepath = findNextFile($AddOns_Path)) > 0; %i++)
	{
		AddAddOnToList(%filepath);
	}
	if (%i < 9)
	{
		echo($AddOns_Amount @ " add-on" @ (($AddOns_Amount > 1) ? "s" : "") @ " found.");
		return;
	}
	schedule(0, 0, "PopulateAddOnList_Schedule");
}

// Add the add-on to a list
function AddAddOnToList(%filepath)
{
	%name = filePath(%filepath);
	%crc = getFileCRC(%name @ ".zip");
	%name = getSubStr(strchr(%name, "/"), 1, strlen(%name));
	
	// This is only for security reasons
	$AddOns_CRC_[%name] = %crc;
	
	// Take only those that haven't been loaded yet
	if ($AddOnLoaded__[%name] != 1)
	{
		$AddOns_Amount++;
		$AddOns_Num_[$AddOns_Amount] = %name;
		$AddOns_Name_[%name] = $AddOns_Amount;
	}
}

// Is a host
function GameConnection::isHost(%client)
{
    return (%client.bl_id == getNumKeyID() || %client.getRawIP() $= "local");
}

if (isPackage(LoadAddOn_Server))
	deactivatePackage(LoadAddOn_Server);

package LoadAddOn_Server
{
	// Transfer list to host
	function GameConnection::onClientEnterGame(%client)
	{
		Parent::onClientEnterGame(%client);
		
		// Host only
		if (!%client.isHost())
			return;
		
		ServerCmdGetAddOns(%client);
	}
	
	// To know if a client is done loading the datablocks
	function GameConnection::onDataBlocksDone(%client, %missionSequence)
	{
		if (!%client.reloadDatablocks)
		{
			Parent::onDataBlocksDone(%client, %missionSequence);
			return;
		}
		%client.reloadDatablocks = false;
		// Reload in a schedule
		%client.schedule(10, onGhostAlwaysObjectsReceived);
	}
};
activatePackage(LoadAddOn_Server);

// Look through all add-ons when everything is loaded to
// prevent findFirstFile to reset making findNextFile not working
schedule(0, 0, PopulateAddOnList);